import math

from PyQt5.QtCore import QPointF, Qt
from PyQt5.QtGui import QPen, QPolygonF
from PyQt5.QtWidgets import QGraphicsItem, QGraphicsTextItem, QGraphicsLineItem

from urh import settings
from urh.simulator.GraphicsItem import GraphicsItem
from urh.simulator.SimulatorMessage import SimulatorMessage
from urh.simulator.UnlabeledRangeItem import UnlabeledRangeItem


class MessageItem(GraphicsItem):
    def __init__(self, model_item: SimulatorMessage, parent=None):
        assert isinstance(model_item, SimulatorMessage)
        super().__init__(model_item=model_item, parent=parent)

        self.setFlag(QGraphicsItem.ItemIsPanel, True)
        self.arrow = MessageArrowItem(self)

        self.repeat_text = QGraphicsTextItem(self)
        self.repeat_text.setFont(self.font)

    def update_flags(self):
        if self.scene().mode == 0:
            self.set_flags(is_selectable=True, is_movable=True, accept_hover_events=True, accept_drops=True)

    def width(self):
        labels = self.labels()
        width = self.number.boundingRect().width()
        # width += 5
        width += sum([lbl.boundingRect().width() for lbl in labels])
        width += 5 * (len(labels) - 1)
        width += self.repeat_text.boundingRect().width()

        return width

    def refresh(self):
        self.repeat_text.setPlainText("(" + str(self.model_item.repeat) + "x)" if self.model_item.repeat > 1 else "")

    def labels(self):
        self.refresh_unlabeled_range_marker()
        unlabeled_range_items = [uri for uri in self.childItems() if isinstance(uri, UnlabeledRangeItem)]
        result = []

        start = 0
        i = 0

        message = self.model_item

        if len(message) and not message.message_type:
            result.append(unlabeled_range_items[0])
        else:
            for lbl in message.message_type:
                if lbl.start > start:
                    result.append(unlabeled_range_items[i])
                    i += 1

                result.append(self.scene().model_to_scene(lbl))
                start = lbl.end

            if start < len(message):
                result.append(unlabeled_range_items[i])

        return result

    def refresh_unlabeled_range_marker(self):
        msg = self.model_item

        urm = [item for item in self.childItems() if isinstance(item, UnlabeledRangeItem)]

        if len(msg):
            num_unlabeled_ranges = len(msg.message_type.unlabeled_ranges)

            if msg.message_type and msg.message_type[-1].end >= len(msg):
                num_unlabeled_ranges -= 1
        else:
            num_unlabeled_ranges = 0

        if len(urm) < num_unlabeled_ranges:
            for i in range(num_unlabeled_ranges - len(urm)):
                UnlabeledRangeItem(self)
        else:
            for i in range(len(urm) - num_unlabeled_ranges):
                self.scene().removeItem(urm[i])

    def update_position(self, x_pos, y_pos):
        labels = self.labels()
        self.setPos(QPointF(x_pos, y_pos))

        p_source = self.mapFromItem(self.source.line, self.source.line.line().p1())
        p_destination = self.mapFromItem(self.destination.line, self.destination.line.line().p1())

        arrow_width = abs(p_source.x() - p_destination.x())

        start_x = min(p_source.x(), p_destination.x())
        start_x += (arrow_width - self.width()) / 2
        start_y = 0

        self.number.setPos(start_x, start_y)
        start_x += self.number.boundingRect().width()

        for label in labels:
            label.setPos(start_x, start_y)
            start_x += label.boundingRect().width() + 5

        self.repeat_text.setPos(start_x, start_y)

        if labels:
            start_y += labels[0].boundingRect().height() + 5
        else:
            start_y += 26

        self.arrow.setLine(p_source.x(), start_y, p_destination.x(), start_y)
        super().update_position(x_pos, y_pos)

    @property
    def source(self):
        return self.scene().participants_dict[self.model_item.participant]

    @property
    def destination(self):
        return self.scene().participants_dict[self.model_item.destination]


class MessageArrowItem(QGraphicsLineItem):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setPen(QPen(Qt.black, 1, Qt.SolidLine, Qt.RoundCap, Qt.RoundJoin))

    def boundingRect(self):
        return super().boundingRect().adjusted(0, -5, 0, 5)

    def paint(self, painter, option, widget):
        if self.line().length() == 0:
            return

        pen = self.pen()
        pen.setColor(settings.LINECOLOR)
        painter.setPen(pen)
        painter.setBrush(settings.LINECOLOR)

        arrow_size = 10.0

        angle = math.acos(self.line().dx() / self.line().length())

        if self.line().dy() >= 0:
            angle = (math.pi * 2) - angle

        arrow_p1 = self.line().p2() - QPointF(math.sin(angle + math.pi / 2.5) * arrow_size,
                                              math.cos(angle + math.pi / 2.5) * arrow_size)

        arrow_p2 = self.line().p2() - QPointF(math.sin(angle + math.pi - math.pi / 2.5) * arrow_size,
                                              math.cos(angle + math.pi - math.pi / 2.5) * arrow_size)

        arrow_head = QPolygonF()
        arrow_head.append(self.line().p2())
        arrow_head.append(arrow_p1)
        arrow_head.append(arrow_p2)

        painter.drawLine(self.line())
        painter.drawPolygon(arrow_head)



import os
import sys
import tempfile

if sys.version_info < (3, 4):
    print("You need at least Python 3.4 for this application!")
    if sys.version_info[0] < 3:
        print("try running with python3 {}".format(" ".join(sys.argv)))
    sys.exit(1)

try:
    from setuptools import setup, Extension
    from setuptools.command.build_ext import build_ext as _build_ext
except ImportError:
    print("Could not find setuptools")
    print("Try installing them with pip install setuptools")
    sys.exit(1)

from src.urh.dev.native import ExtensionHelper
from src.urh.dev.native.ExtensionHelper import COMPILER_DIRECTIVES
import src.urh.version as version

if sys.platform == "win32":
    OPEN_MP_FLAG = "/openmp"
    NO_NUMPY_WARNINGS_FLAG = ""
elif sys.platform == "darwin":
    OPEN_MP_FLAG = ""  # no OpenMP support in default Mac OSX compiler
    NO_NUMPY_WARNINGS_FLAG = "-Wno-#warnings"
else:
    OPEN_MP_FLAG = "-fopenmp"
    NO_NUMPY_WARNINGS_FLAG = "-Wno-cpp"

UI_SUBDIRS = ("actions", "delegates", "views")
PLUGINS = [path for path in os.listdir("src/urh/plugins") if os.path.isdir(os.path.join("src/urh/plugins", path))]
URH_DIR = "urh"

IS_RELEASE = os.path.isfile(os.path.join(tempfile.gettempdir(), "urh_releasing"))

try:
    from Cython.Build import cythonize
except ImportError:
    print("You need Cython to build URH's extensions!\n"
          "You can get it e.g. with python3 -m pip install cython.",
          file=sys.stderr)
    sys.exit(1)


class build_ext(_build_ext):
    def finalize_options(self):
        print("Finalizing options")
        _build_ext.finalize_options(self)
        # Prevent numpy from thinking it is still in its setup process:
        __builtins__.__NUMPY_SETUP__ = False
        import numpy
        self.include_dirs.append(numpy.get_include())


def get_packages():
    packages = [URH_DIR]
    separator = os.path.normpath("/")
    for dirpath, dirnames, filenames in os.walk(os.path.join("./src/", URH_DIR)):
        package_path = os.path.relpath(dirpath, os.path.join("./src/", URH_DIR)).replace(separator, ".")
        if len(package_path) > 1:
            packages.append(URH_DIR + "." + package_path)

    return packages


def get_package_data():
    package_data = {"urh.cythonext": ["*.pyx", "*.pxd"]}
    for plugin in PLUGINS:
        package_data["urh.plugins." + plugin] = ['*.ui', "*.txt"]

    package_data["urh.dev.native.lib"] = ["*.pyx", "*.pxd"]

    if IS_RELEASE and sys.platform == "win32":
        package_data["urh.dev.native.lib.shared"] = ["*.dll", "*.txt"]

    return package_data


def get_extensions():
    filenames = [os.path.splitext(f)[0] for f in os.listdir("src/urh/cythonext") if f.endswith(".pyx")]
    extensions = [Extension("urh.cythonext." + f, ["src/urh/cythonext/" + f + ".pyx"],
                            extra_compile_args=[OPEN_MP_FLAG],
                            extra_link_args=[OPEN_MP_FLAG],
                            language="c++") for f in filenames]

    ExtensionHelper.USE_RELATIVE_PATHS = True
    device_extensions, device_extras = ExtensionHelper.get_device_extensions_and_extras()
    extensions += device_extensions

    if NO_NUMPY_WARNINGS_FLAG:
        for extension in extensions:
            extension.extra_compile_args.append(NO_NUMPY_WARNINGS_FLAG)

    extensions = cythonize(extensions, compiler_directives=COMPILER_DIRECTIVES, compile_time_env=device_extras)
    return extensions


def read_long_description():
    try:
        with open("README.md") as f:
            text = f.read()
        return text
    except:
        return ""


install_requires = ["numpy", "psutil", "cython"]
if IS_RELEASE:
    install_requires.append("pyqt5")
else:
    try:
        import PyQt5
    except ImportError:
        install_requires.append("pyqt5")

if sys.version_info < (3, 4):
    install_requires.append('enum34')

setup(
    name="urh",
    version=version.VERSION,
    description="Universal Radio Hacker: investigate wireless protocols like a boss",
    long_description=read_long_description(),
    long_description_content_type="text/markdown",
    author="Johannes Pohl",
    author_email="Johannes.Pohl90@gmail.com",
    package_dir={"": "src"},
    package_data=get_package_data(),
    url="https://github.com/jopohl/urh",
    license="GNU General Public License (GPL)",
    download_url="https://github.com/jopohl/urh/tarball/v" + str(version.VERSION),
    install_requires=install_requires,
    setup_requires=['numpy'],
    packages=get_packages(),
    ext_modules=get_extensions(),
    cmdclass={'build_ext': build_ext},
    zip_safe=False,
    entry_points={
        'console_scripts': [
            'urh = urh.main:main',
            'urh_cli = urh.cli.urh_cli:main',
        ]}
)



import copy

from PyQt5.QtWidgets import QUndoCommand

from urh.signalprocessing.ProtocolAnalyzer import ProtocolAnalyzer


class InsertColumn(QUndoCommand):
    def __init__(self, proto_analyzer: ProtocolAnalyzer, index: int, rows: list, view: int):
        super().__init__()
        self.proto_analyzer = proto_analyzer
        self.index = proto_analyzer.convert_index(index, from_view=view, to_view=0, decoded=False)[0]
        self.nbits = 1 if view == 0 else 4 if view == 1 else 8
        self.rows = rows

        self.saved_messages = {}

        self.setText("Insert column at {0:d}".format(index))

    def redo(self):
        for i in self.rows:
            msg = self.proto_analyzer.messages[i]
            self.saved_messages[i] = copy.deepcopy(msg)
            for j in range(self.nbits):
                msg.insert(self.index + j, False)

    def undo(self):
        for i in self.rows:
            self.proto_analyzer.messages[i] = self.saved_messages[i]
        self.saved_messages.clear()



from cplutosdr cimport *
from libc.stdint cimport int16_t

from libc.stdlib cimport malloc, free
cdef iio_context* _c_context


cdef iio_buffer* _rx_buffer
cdef iio_buffer* _tx_buffer

cdef bool IS_TX = True
cdef size_t RX_BUFFER_SIZE = 4096

cpdef set_tx(bool is_tx):
    global IS_TX
    IS_TX = <bool>is_tx

cpdef bool get_tx():
    return IS_TX

cpdef tuple scan_devices():
    cdef iio_scan_context* ctx = iio_create_scan_context(NULL, 0)

    cdef iio_context_info** infos
    cdef ssize_t num_devices = iio_scan_context_get_info_list(ctx, &infos)

    cdef list descs = []
    cdef list uris = []
    for i in range(0, num_devices):
        desc = iio_context_info_get_description(infos[0])
        uri = iio_context_info_get_uri(infos[0])
        descs.append(desc.decode())
        uris.append(uri.decode())

    iio_context_info_list_free(infos)
    iio_scan_context_destroy(ctx)

    return descs, uris

cpdef int set_center_freq(long long center_freq):
    cdef iio_device* phy = iio_context_find_device(_c_context, "ad9361-phy")
    if IS_TX:
        return iio_channel_attr_write_longlong(iio_device_find_channel(phy, "altvoltage1", True), "frequency", center_freq)
    else:
        return iio_channel_attr_write_longlong(iio_device_find_channel(phy, "altvoltage0", True), "frequency", center_freq)

cpdef int set_sample_rate(long long sample_rate):
    cdef iio_device* phy = iio_context_find_device(_c_context, "ad9361-phy")
    return iio_channel_attr_write_longlong(iio_device_find_channel(phy, "voltage0", IS_TX), "sampling_frequency", sample_rate)

cpdef int set_bandwidth(long long bandwidth):
    cdef iio_device* phy = iio_context_find_device(_c_context, "ad9361-phy")
    return iio_channel_attr_write_longlong(iio_device_find_channel(phy, "voltage0", IS_TX), "rf_bandwidth", bandwidth)


cpdef int set_rf_gain(long long gain):
    cdef iio_device* phy = iio_context_find_device(_c_context, "ad9361-phy")
    iio_channel_attr_write(iio_device_find_channel(phy, "voltage0", IS_TX), "gain_control_mode", "manual")
    return iio_channel_attr_write_longlong(iio_device_find_channel(phy, "voltage0", IS_TX), "hardwaregain", gain)

cpdef int open(str uri):
    global _c_context
    _c_context = iio_create_context_from_uri(uri.encode())

    if _c_context != NULL:
        return 0
    else:
        return -1

cpdef int setup_rx(size_t buffer_size):
    dev = iio_context_find_device(_c_context, "cf-ad9361-lpc")
    set_rx_channels_status(enable=True)
    global _rx_buffer

    global RX_BUFFER_SIZE
    RX_BUFFER_SIZE = buffer_size
    _rx_buffer = iio_device_create_buffer(dev, RX_BUFFER_SIZE, False)
    return 0

cpdef int setup_tx(size_t buffer_size):
    dev = iio_context_find_device(_c_context, "cf-ad9361-dds-core-lpc")
    set_tx_channels_status(enable=True)
    global _tx_buffer
    _tx_buffer = iio_device_create_buffer(dev, buffer_size, False)
    return 0

cpdef bytes receive_sync(connection):
    cdef iio_device* dev = iio_context_find_device(_c_context, "cf-ad9361-lpc")
    cdef char *p_dat = <char*>iio_buffer_first(_rx_buffer, iio_device_find_channel(dev, "voltage0", False))
    cdef char *p_end = <char*>iio_buffer_end(_rx_buffer)
    cdef ssize_t p_inc = iio_buffer_step(_rx_buffer)
    cdef int16_t i, q

    iio_buffer_refill(_rx_buffer)

    cdef int16_t *samples = <int16_t *> malloc(2*RX_BUFFER_SIZE * sizeof(int16_t))
    cdef unsigned int index = 0

    try:
        while p_dat < p_end:
            i = (<int16_t*>p_dat)[0]
            q = (<int16_t*>p_dat)[1]

            samples[index] = i
            samples[index+1] = q
            index += 2
            p_dat += p_inc

        connection.send_bytes(<int16_t [:2*RX_BUFFER_SIZE]>samples)
    finally:
        free(samples)

cpdef int send_sync(int16_t[::1] samples):
    cdef unsigned int i = 0
    cdef iio_device* dev = iio_context_find_device(_c_context, "cf-ad9361-dds-core-lpc")
    cdef char *p_dat = <char*>iio_buffer_first(_tx_buffer, iio_device_find_channel(dev, "voltage0", True))
    cdef char *p_end = <char*>iio_buffer_end(_tx_buffer)
    cdef ssize_t p_inc = iio_buffer_step(_tx_buffer)

    cdef unsigned int n = len(samples)

    while p_dat < p_end:
        if i < n:
            (<int16_t*>p_dat)[0] = samples[i]
        else:
            (<int16_t*>p_dat)[0] = 0

        if i + 1 < n:
            (<int16_t*>p_dat)[1] = samples[i+1]
        else:
            (<int16_t*>p_dat)[0] = 0

        p_dat += p_inc
        i += 2

    n_send = iio_buffer_push(_tx_buffer)
    return 0 if n_send > 0 else -1

cpdef stop_rx():
    set_rx_channels_status(enable=False)
    iio_buffer_destroy(_rx_buffer)
    return 0

cpdef stop_tx():
    set_tx_channels_status(enable=False)
    iio_buffer_destroy(_tx_buffer)
    return 0

cpdef int close():
    if IS_TX:
        stop_tx()
    else:
        stop_rx()

    iio_context_destroy(_c_context)

cpdef set_rx_channels_status(bool enable):
    cdef iio_channel *rx0_i
    cdef iio_channel *rx0_q
    cdef iio_device* dev = iio_context_find_device(_c_context, "cf-ad9361-lpc")
    rx0_i = iio_device_find_channel(dev, "voltage0", False)
    rx0_q = iio_device_find_channel(dev, "voltage1", False)

    if enable:
        iio_channel_enable(rx0_i)
        iio_channel_enable(rx0_q)
    else:
        iio_channel_disable(rx0_i)
        iio_channel_disable(rx0_q)


cpdef set_tx_channels_status(bool enable):
    cdef iio_channel *tx0_i
    cdef iio_channel *tx0_q
    cdef iio_device* dev = iio_context_find_device(_c_context, "cf-ad9361-dds-core-lpc")
    tx0_i = iio_device_find_channel(dev, "voltage0", True)
    tx0_q = iio_device_find_channel(dev, "voltage1", True)

    if enable:
        iio_channel_enable(tx0_i)
        iio_channel_enable(tx0_q)
    else:
        iio_channel_disable(tx0_i)
        iio_channel_disable(tx0_q)



import copy

from PyQt5.QtWidgets import QUndoCommand

from urh.signalprocessing.ProtocolAnalyzer import ProtocolAnalyzer
from urh.signalprocessing.Message import Message


class MessageBreakAction(QUndoCommand):
    def __init__(self, proto_analyzer: ProtocolAnalyzer, msg_nr: int, pos: int):
        super().__init__()
        self.proto_analyzer = proto_analyzer
        self.msg_nr = msg_nr
        self.pos = pos
        self.orig_messages = copy.deepcopy(proto_analyzer.messages)

        self.setText("Break message behind selection")

    def redo(self):
        message = self.proto_analyzer.messages[self.msg_nr]
        message1 = Message(plain_bits=message.plain_bits[:self.pos], pause=0,
                           rssi=message.rssi, decoder=message.decoder, message_type=message.message_type,
                           samples_per_symbol=message.samples_per_symbol)
        message2 = Message(plain_bits=message.plain_bits[self.pos:], pause=message.pause,
                           rssi=message.rssi, decoder=message.decoder, message_type=message.message_type,
                           samples_per_symbol=message.samples_per_symbol)
        self.proto_analyzer.messages[self.msg_nr] = message1
        self.proto_analyzer.messages.insert(self.msg_nr + 1, message2)

    def undo(self):
        self.proto_analyzer.messages = self.orig_messages

    def __get_zero_seq_indexes(self, message: str, following_zeros: int):
        """
        :rtype: list[tuple of int]
        """

        result = []
        if following_zeros > len(message):
            return result

        zero_counter = 0
        for i in range(0, len(message)):
            if message[i] == "0":
                zero_counter += 1
            else:
                if zero_counter >= following_zeros:
                    result.append((i - zero_counter, i))
                zero_counter = 0

        if zero_counter >= following_zeros:
            result.append((len(message) - 1 - following_zeros, len(message) - 1))

        return result



import math

import numpy as np
from PyQt5.QtCore import QRectF
from PyQt5.QtGui import QPen, QFont, QTransform, QFontMetrics
from PyQt5.QtWidgets import QGraphicsScene, QGraphicsRectItem, QGraphicsSceneDragDropEvent, QGraphicsSimpleTextItem

from urh import settings
from urh.ui.painting.HorizontalSelection import HorizontalSelection
from urh.util import util


class ZoomableScene(QGraphicsScene):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.noise_area = None

        self.separation_areas = []  # type: list[QGraphicsRectItem]
        self.captions = []  # type: list[QGraphicsSimpleTextItem]

        self.centers = [0]

        self.always_show_symbols_legend = False

        self.ones_caption = None
        self.zeros_caption = None

        self.ones_arrow = None
        self.zeros_arrow = None
        self.selection_area = HorizontalSelection(0, 0, 0, 0, fillcolor=settings.SELECTION_COLOR,
                                                  opacity=settings.SELECTION_OPACITY)
        self.addItem(self.selection_area)

    @property
    def bits_per_symbol(self):
        return int(math.log2(len(self.centers) + 1))

    def draw_noise_area(self, y, h):
        x = self.sceneRect().x()
        w = self.sceneRect().width()

        for area in self.separation_areas:
            area.hide()

        if self.noise_area is None or self.noise_area.scene() != self:
            roi = HorizontalSelection(x, y, w, h, fillcolor=settings.NOISE_COLOR, opacity=settings.NOISE_OPACITY)
            self.noise_area = roi
            self.addItem(self.noise_area)
        else:
            self.noise_area.show()
            self.noise_area.setY(y)
            self.noise_area.height = h

    def hide_legend(self):
        for caption in self.captions:
            caption.hide()

    def clear_legend(self):
        for caption in self.captions:
            self.removeItem(caption)
        self.captions.clear()

    def redraw_legend(self, force_show=False):
        if not (force_show or self.always_show_symbols_legend):
            self.hide_legend()
            return

        num_captions = len(self.centers) + 1
        if num_captions != len(self.captions):
            self.clear_legend()

            fmt = "{0:0" + str(self.bits_per_symbol) + "b}"
            for i in range(num_captions):
                font = QFont()
                font.setPointSize(16)
                font.setBold(True)
                self.captions.append(self.addSimpleText(fmt.format(i), font))

        view_rect = self.parent().view_rect()  # type: QRectF
        padding = 0
        fm = QFontMetrics(self.captions[0].font())

        for i, caption in enumerate(self.captions):
            caption.show()
            scale_x, scale_y = util.calc_x_y_scale(self.separation_areas[i].rect(), self.parent())
            try:
                caption.setPos(view_rect.x() + view_rect.width() - fm.width(caption.text()) * scale_x,
                               self.centers[i] + padding)
            except IndexError:
                caption.setPos(view_rect.x() + view_rect.width() - fm.width(caption.text()) * scale_x,
                               self.centers[i - 1] - padding - fm.height() * scale_y)

            caption.setTransform(QTransform.fromScale(scale_x, scale_y), False)

    def draw_sep_area(self, centers: np.ndarray, show_symbols=False):
        x = self.sceneRect().x()
        y = self.sceneRect().y()
        w = self.sceneRect().width()
        h = self.sceneRect().height()
        reversed_centers = list(reversed(centers))

        num_areas = len(centers) + 1
        if num_areas != len(self.separation_areas):
            for area in self.separation_areas:
                self.removeItem(area)
            self.separation_areas.clear()

            for i in range(num_areas):
                area = QGraphicsRectItem(0, 0, 0, 0)
                if i % 2 == 0:
                    area.setBrush(settings.ZEROS_AREA_COLOR)
                else:
                    area.setBrush(settings.ONES_AREA_COLOR)
                area.setOpacity(settings.SEPARATION_OPACITY)
                area.setPen(QPen(settings.TRANSPARENT_COLOR, 0))
                self.addItem(area)
                self.separation_areas.append(area)

        start = y

        for i, area in enumerate(self.separation_areas):
            area.show()
            try:
                self.separation_areas[i].setRect(x, start, w, abs(start - reversed_centers[i]))
                start += abs(start - reversed_centers[i])
            except IndexError:
                self.separation_areas[i].setRect(x, start, w, abs(start - h))

        if self.noise_area is not None:
            self.noise_area.hide()

        self.centers = centers
        self.redraw_legend(show_symbols)

    def clear(self):
        self.noise_area = None
        self.separation_areas.clear()
        self.captions.clear()
        self.selection_area = None
        super().clear()

    def dragEnterEvent(self, event: QGraphicsSceneDragDropEvent):
        event.accept()

    def dragMoveEvent(self, event: QGraphicsSceneDragDropEvent):
        event.accept()



# -*- coding: utf-8 -*-

#
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_DialogOptions(object):
    def setupUi(self, DialogOptions):
        DialogOptions.setObjectName("DialogOptions")
        DialogOptions.resize(814, 822)
        icon = QtGui.QIcon.fromTheme("configure")
        DialogOptions.setWindowIcon(icon)
        self.verticalLayout_6 = QtWidgets.QVBoxLayout(DialogOptions)
        self.verticalLayout_6.setObjectName("verticalLayout_6")
        self.tabWidget = QtWidgets.QTabWidget(DialogOptions)
        self.tabWidget.setObjectName("tabWidget")
        self.tabGeneration = QtWidgets.QWidget()
        self.tabGeneration.setObjectName("tabGeneration")
        self.verticalLayout_9 = QtWidgets.QVBoxLayout(self.tabGeneration)
        self.verticalLayout_9.setObjectName("verticalLayout_9")
        self.gridLayout_4 = QtWidgets.QGridLayout()
        self.gridLayout_4.setObjectName("gridLayout_4")
        self.labelFuzzingSamples = QtWidgets.QLabel(self.tabGeneration)
        self.labelFuzzingSamples.setObjectName("labelFuzzingSamples")
        self.gridLayout_4.addWidget(self.labelFuzzingSamples, 1, 1, 1, 1)
        self.checkBoxDefaultFuzzingPause = QtWidgets.QCheckBox(self.tabGeneration)
        self.checkBoxDefaultFuzzingPause.setObjectName("checkBoxDefaultFuzzingPause")
        self.gridLayout_4.addWidget(self.checkBoxDefaultFuzzingPause, 0, 0, 1, 2)
        self.doubleSpinBoxFuzzingPause = KillerDoubleSpinBox(self.tabGeneration)
        self.doubleSpinBoxFuzzingPause.setDecimals(3)
        self.doubleSpinBoxFuzzingPause.setMaximum(999999999.0)
        self.doubleSpinBoxFuzzingPause.setObjectName("doubleSpinBoxFuzzingPause")
        self.gridLayout_4.addWidget(self.doubleSpinBoxFuzzingPause, 1, 0, 1, 1)
        self.checkBoxMultipleModulations = QtWidgets.QCheckBox(self.tabGeneration)
        self.checkBoxMultipleModulations.setObjectName("checkBoxMultipleModulations")
        self.gridLayout_4.addWidget(self.checkBoxMultipleModulations, 2, 0, 1, 2)
        self.verticalLayout_9.addLayout(self.gridLayout_4)
        self.groupBoxModulationAccuracy = QtWidgets.QGroupBox(self.tabGeneration)
        self.groupBoxModulationAccuracy.setObjectName("groupBoxModulationAccuracy")
        self.verticalLayout_7 = QtWidgets.QVBoxLayout(self.groupBoxModulationAccuracy)
        self.verticalLayout_7.setObjectName("verticalLayout_7")
        self.radioButtonLowModulationAccuracy = QtWidgets.QRadioButton(self.groupBoxModulationAccuracy)
        self.radioButtonLowModulationAccuracy.setObjectName("radioButtonLowModulationAccuracy")
        self.verticalLayout_7.addWidget(self.radioButtonLowModulationAccuracy)
        self.radioButtonMediumModulationAccuracy = QtWidgets.QRadioButton(self.groupBoxModulationAccuracy)
        self.radioButtonMediumModulationAccuracy.setObjectName("radioButtonMediumModulationAccuracy")
        self.verticalLayout_7.addWidget(self.radioButtonMediumModulationAccuracy)
        self.radioButtonHighModulationAccuracy = QtWidgets.QRadioButton(self.groupBoxModulationAccuracy)
        self.radioButtonHighModulationAccuracy.setObjectName("radioButtonHighModulationAccuracy")
        self.verticalLayout_7.addWidget(self.radioButtonHighModulationAccuracy)
        self.verticalLayout_9.addWidget(self.groupBoxModulationAccuracy)
        spacerItem = QtWidgets.QSpacerItem(20, 500, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_9.addItem(spacerItem)
        self.tabWidget.addTab(self.tabGeneration, "")
        self.tabView = QtWidgets.QWidget()
        self.tabView.setObjectName("tabView")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.tabView)
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.label_7 = QtWidgets.QLabel(self.tabView)
        self.label_7.setObjectName("label_7")
        self.horizontalLayout_2.addWidget(self.label_7)
        self.comboBoxDefaultView = QtWidgets.QComboBox(self.tabView)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.comboBoxDefaultView.sizePolicy().hasHeightForWidth())
        self.comboBoxDefaultView.setSizePolicy(sizePolicy)
        self.comboBoxDefaultView.setObjectName("comboBoxDefaultView")
        self.comboBoxDefaultView.addItem("")
        self.comboBoxDefaultView.addItem("")
        self.comboBoxDefaultView.addItem("")
        self.horizontalLayout_2.addWidget(self.comboBoxDefaultView)
        self.verticalLayout.addLayout(self.horizontalLayout_2)
        self.checkBoxShowConfirmCloseDialog = QtWidgets.QCheckBox(self.tabView)
        self.checkBoxShowConfirmCloseDialog.setObjectName("checkBoxShowConfirmCloseDialog")
        self.verticalLayout.addWidget(self.checkBoxShowConfirmCloseDialog)
        self.checkBoxHoldShiftToDrag = QtWidgets.QCheckBox(self.tabView)
        self.checkBoxHoldShiftToDrag.setObjectName("checkBoxHoldShiftToDrag")
        self.verticalLayout.addWidget(self.checkBoxHoldShiftToDrag)
        self.checkBoxPauseTime = QtWidgets.QCheckBox(self.tabView)
        self.checkBoxPauseTime.setObjectName("checkBoxPauseTime")
        self.verticalLayout.addWidget(self.checkBoxPauseTime)
        self.checkBoxAlignLabels = QtWidgets.QCheckBox(self.tabView)
        self.checkBoxAlignLabels.setObjectName("checkBoxAlignLabels")
        self.verticalLayout.addWidget(self.checkBoxAlignLabels)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.labelFontSize = QtWidgets.QLabel(self.tabView)
        self.labelFontSize.setObjectName("labelFontSize")
        self.horizontalLayout.addWidget(self.labelFontSize)
        self.doubleSpinBoxFontSize = QtWidgets.QDoubleSpinBox(self.tabView)
        self.doubleSpinBoxFontSize.setMinimum(1.0)
        self.doubleSpinBoxFontSize.setObjectName("doubleSpinBoxFontSize")
        self.horizontalLayout.addWidget(self.doubleSpinBoxFontSize)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.label_9 = QtWidgets.QLabel(self.tabView)
        self.label_9.setObjectName("label_9")
        self.horizontalLayout_4.addWidget(self.label_9)
        self.comboBoxTheme = QtWidgets.QComboBox(self.tabView)
        self.comboBoxTheme.setObjectName("comboBoxTheme")
        self.comboBoxTheme.addItem("")
        self.comboBoxTheme.addItem("")
        self.comboBoxTheme.addItem("")
        self.horizontalLayout_4.addWidget(self.comboBoxTheme)
        self.verticalLayout.addLayout(self.horizontalLayout_4)
        self.horizontalLayout_5 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_5.setObjectName("horizontalLayout_5")
        self.labelIconTheme = QtWidgets.QLabel(self.tabView)
        self.labelIconTheme.setObjectName("labelIconTheme")
        self.horizontalLayout_5.addWidget(self.labelIconTheme)
        self.comboBoxIconTheme = QtWidgets.QComboBox(self.tabView)
        self.comboBoxIconTheme.setObjectName("comboBoxIconTheme")
        self.comboBoxIconTheme.addItem("")
        self.comboBoxIconTheme.addItem("")
        self.horizontalLayout_5.addWidget(self.comboBoxIconTheme)
        self.verticalLayout.addLayout(self.horizontalLayout_5)
        self.groupBoxSpectrogramColormap = QtWidgets.QGroupBox(self.tabView)
        self.groupBoxSpectrogramColormap.setObjectName("groupBoxSpectrogramColormap")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.groupBoxSpectrogramColormap)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.scrollAreaSpectrogramColormap = QtWidgets.QScrollArea(self.groupBoxSpectrogramColormap)
        self.scrollAreaSpectrogramColormap.setWidgetResizable(True)
        self.scrollAreaSpectrogramColormap.setObjectName("scrollAreaSpectrogramColormap")
        self.scrollAreaWidgetSpectrogramColormapContents = QtWidgets.QWidget()
        self.scrollAreaWidgetSpectrogramColormapContents.setGeometry(QtCore.QRect(0, 0, 762, 466))
        self.scrollAreaWidgetSpectrogramColormapContents.setObjectName("scrollAreaWidgetSpectrogramColormapContents")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(self.scrollAreaWidgetSpectrogramColormapContents)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.scrollAreaSpectrogramColormap.setWidget(self.scrollAreaWidgetSpectrogramColormapContents)
        self.verticalLayout_2.addWidget(self.scrollAreaSpectrogramColormap)
        self.verticalLayout.addWidget(self.groupBoxSpectrogramColormap)
        self.tabWidget.addTab(self.tabView, "")
        self.tabFieldtypes = QtWidgets.QWidget()
        self.tabFieldtypes.setObjectName("tabFieldtypes")
        self.verticalLayout_5 = QtWidgets.QVBoxLayout(self.tabFieldtypes)
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.tblLabeltypes = QtWidgets.QTableView(self.tabFieldtypes)
        self.tblLabeltypes.setAlternatingRowColors(True)
        self.tblLabeltypes.setObjectName("tblLabeltypes")
        self.horizontalLayout_3.addWidget(self.tblLabeltypes)
        self.verticalLayout_3 = QtWidgets.QVBoxLayout()
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.btnAddLabelType = QtWidgets.QToolButton(self.tabFieldtypes)
        icon = QtGui.QIcon.fromTheme("list-add")
        self.btnAddLabelType.setIcon(icon)
        self.btnAddLabelType.setObjectName("btnAddLabelType")
        self.verticalLayout_3.addWidget(self.btnAddLabelType)
        self.btnRemoveLabeltype = QtWidgets.QToolButton(self.tabFieldtypes)
        icon = QtGui.QIcon.fromTheme("list-remove")
        self.btnRemoveLabeltype.setIcon(icon)
        self.btnRemoveLabeltype.setObjectName("btnRemoveLabeltype")
        self.verticalLayout_3.addWidget(self.btnRemoveLabeltype)
        spacerItem1 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_3.addItem(spacerItem1)
        self.horizontalLayout_3.addLayout(self.verticalLayout_3)
        self.verticalLayout_5.addLayout(self.horizontalLayout_3)
        spacerItem2 = QtWidgets.QSpacerItem(20, 203, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_5.addItem(spacerItem2)
        self.tabWidget.addTab(self.tabFieldtypes, "")
        self.tab_plugins = QtWidgets.QWidget()
        self.tab_plugins.setObjectName("tab_plugins")
        self.tabWidget.addTab(self.tab_plugins, "")
        self.tabDevices = QtWidgets.QWidget()
        self.tabDevices.setObjectName("tabDevices")
        self.verticalLayout_8 = QtWidgets.QVBoxLayout(self.tabDevices)
        self.verticalLayout_8.setObjectName("verticalLayout_8")
        self.labelInfoDeviceTable = QtWidgets.QLabel(self.tabDevices)
        font = QtGui.QFont()
        font.setItalic(True)
        self.labelInfoDeviceTable.setFont(font)
        self.labelInfoDeviceTable.setToolTip("")
        self.labelInfoDeviceTable.setWordWrap(True)
        self.labelInfoDeviceTable.setObjectName("labelInfoDeviceTable")
        self.verticalLayout_8.addWidget(self.labelInfoDeviceTable)
        self.tblDevices = QtWidgets.QTableView(self.tabDevices)
        self.tblDevices.setAlternatingRowColors(True)
        self.tblDevices.setShowGrid(False)
        self.tblDevices.setObjectName("tblDevices")
        self.tblDevices.horizontalHeader().setDefaultSectionSize(200)
        self.tblDevices.verticalHeader().setVisible(False)
        self.verticalLayout_8.addWidget(self.tblDevices)
        self.labelDeviceMissingInfo = QtWidgets.QLabel(self.tabDevices)
        font = QtGui.QFont()
        font.setItalic(False)
        self.labelDeviceMissingInfo.setFont(font)
        self.labelDeviceMissingInfo.setWordWrap(True)
        self.labelDeviceMissingInfo.setObjectName("labelDeviceMissingInfo")
        self.verticalLayout_8.addWidget(self.labelDeviceMissingInfo)
        self.line = QtWidgets.QFrame(self.tabDevices)
        self.line.setFrameShape(QtWidgets.QFrame.HLine)
        self.line.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line.setObjectName("line")
        self.verticalLayout_8.addWidget(self.line)
        self.groupBoxNativeOptions = QtWidgets.QGroupBox(self.tabDevices)
        self.groupBoxNativeOptions.setObjectName("groupBoxNativeOptions")
        self.gridLayout_5 = QtWidgets.QGridLayout(self.groupBoxNativeOptions)
        self.gridLayout_5.setObjectName("gridLayout_5")
        self.labelLibDirs = QtWidgets.QLabel(self.groupBoxNativeOptions)
        self.labelLibDirs.setObjectName("labelLibDirs")
        self.gridLayout_5.addWidget(self.labelLibDirs, 2, 0, 1, 1)
        self.lineEditIncludeDirs = QtWidgets.QLineEdit(self.groupBoxNativeOptions)
        self.lineEditIncludeDirs.setObjectName("lineEditIncludeDirs")
        self.gridLayout_5.addWidget(self.lineEditIncludeDirs, 3, 2, 1, 1)
        self.labelNativeRebuildInfo = QtWidgets.QLabel(self.groupBoxNativeOptions)
        self.labelNativeRebuildInfo.setWordWrap(True)
        self.labelNativeRebuildInfo.setObjectName("labelNativeRebuildInfo")
        self.gridLayout_5.addWidget(self.labelNativeRebuildInfo, 1, 0, 1, 3)
        self.lineEditLibDirs = QtWidgets.QLineEdit(self.groupBoxNativeOptions)
        self.lineEditLibDirs.setObjectName("lineEditLibDirs")
        self.gridLayout_5.addWidget(self.lineEditLibDirs, 2, 2, 1, 1)
        self.labelIncludeDirs = QtWidgets.QLabel(self.groupBoxNativeOptions)
        self.labelIncludeDirs.setObjectName("labelIncludeDirs")
        self.gridLayout_5.addWidget(self.labelIncludeDirs, 3, 0, 1, 1)
        self.btnViewBuildLog = QtWidgets.QPushButton(self.groupBoxNativeOptions)
        icon = QtGui.QIcon.fromTheme("utilities-log-viewer")
        self.btnViewBuildLog.setIcon(icon)
        self.btnViewBuildLog.setObjectName("btnViewBuildLog")
        self.gridLayout_5.addWidget(self.btnViewBuildLog, 4, 1, 1, 1)
        self.btnRebuildNative = QtWidgets.QPushButton(self.groupBoxNativeOptions)
        self.btnRebuildNative.setEnabled(True)
        icon = QtGui.QIcon.fromTheme("view-refresh")
        self.btnRebuildNative.setIcon(icon)
        self.btnRebuildNative.setObjectName("btnRebuildNative")
        self.gridLayout_5.addWidget(self.btnRebuildNative, 4, 0, 1, 1)
        self.labelRebuildNativeStatus = QtWidgets.QLabel(self.groupBoxNativeOptions)
        self.labelRebuildNativeStatus.setAlignment(QtCore.Qt.AlignLeading|QtCore.Qt.AlignLeft|QtCore.Qt.AlignVCenter)
        self.labelRebuildNativeStatus.setObjectName("labelRebuildNativeStatus")
        self.gridLayout_5.addWidget(self.labelRebuildNativeStatus, 4, 2, 1, 1)
        self.verticalLayout_8.addWidget(self.groupBoxNativeOptions)
        self.line_2 = QtWidgets.QFrame(self.tabDevices)
        self.line_2.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_2.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_2.setObjectName("line_2")
        self.verticalLayout_8.addWidget(self.line_2)
        self.groupBox_3 = QtWidgets.QGroupBox(self.tabDevices)
        self.groupBox_3.setObjectName("groupBox_3")
        self.gridLayout_2 = QtWidgets.QGridLayout(self.groupBox_3)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.label = QtWidgets.QLabel(self.groupBox_3)
        self.label.setObjectName("label")
        self.gridLayout_2.addWidget(self.label, 0, 0, 1, 1)
        self.lineEditGRPythonInterpreter = QtWidgets.QLineEdit(self.groupBox_3)
        self.lineEditGRPythonInterpreter.setObjectName("lineEditGRPythonInterpreter")
        self.gridLayout_2.addWidget(self.lineEditGRPythonInterpreter, 0, 1, 1, 1)
        self.btnChooseGRPythonInterpreter = QtWidgets.QToolButton(self.groupBox_3)
        self.btnChooseGRPythonInterpreter.setObjectName("btnChooseGRPythonInterpreter")
        self.gridLayout_2.addWidget(self.btnChooseGRPythonInterpreter, 0, 2, 1, 1)
        self.verticalLayout_8.addWidget(self.groupBox_3)
        self.line_3 = QtWidgets.QFrame(self.tabDevices)
        self.line_3.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_3.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_3.setObjectName("line_3")
        self.verticalLayout_8.addWidget(self.line_3)
        self.gridLayout_3 = QtWidgets.QGridLayout()
        self.gridLayout_3.setObjectName("gridLayout_3")
        self.label_8 = QtWidgets.QLabel(self.tabDevices)
        self.label_8.setObjectName("label_8")
        self.gridLayout_3.addWidget(self.label_8, 0, 0, 1, 1)
        self.spinBoxNumSendingRepeats = QtWidgets.QSpinBox(self.tabDevices)
        self.spinBoxNumSendingRepeats.setProperty("showGroupSeparator", False)
        self.spinBoxNumSendingRepeats.setMaximum(999999999)
        self.spinBoxNumSendingRepeats.setDisplayIntegerBase(10)
        self.spinBoxNumSendingRepeats.setObjectName("spinBoxNumSendingRepeats")
        self.gridLayout_3.addWidget(self.spinBoxNumSendingRepeats, 0, 1, 1, 1)
        self.label_5 = QtWidgets.QLabel(self.tabDevices)
        self.label_5.setObjectName("label_5")
        self.gridLayout_3.addWidget(self.label_5, 1, 0, 1, 1)
        self.doubleSpinBoxRAMThreshold = QtWidgets.QDoubleSpinBox(self.tabDevices)
        self.doubleSpinBoxRAMThreshold.setMinimum(1.0)
        self.doubleSpinBoxRAMThreshold.setMaximum(100.0)
        self.doubleSpinBoxRAMThreshold.setObjectName("doubleSpinBoxRAMThreshold")
        self.gridLayout_3.addWidget(self.doubleSpinBoxRAMThreshold, 1, 1, 1, 1)
        self.verticalLayout_8.addLayout(self.gridLayout_3)
        self.tabWidget.addTab(self.tabDevices, "")
        self.verticalLayout_6.addWidget(self.tabWidget)

        self.retranslateUi(DialogOptions)
        self.tabWidget.setCurrentIndex(4)

    def retranslateUi(self, DialogOptions):
        _translate = QtCore.QCoreApplication.translate
        DialogOptions.setWindowTitle(_translate("DialogOptions", "Options"))
        self.labelFuzzingSamples.setText(_translate("DialogOptions", "Samples"))
        self.checkBoxDefaultFuzzingPause.setToolTip(_translate("DialogOptions", "<html><head/><body><p>If you disable the default pause, the pause of the fuzzed message will be used.</p></body></html>"))
        self.checkBoxDefaultFuzzingPause.setText(_translate("DialogOptions", "Use a default pause for fuzzed messages"))
        self.checkBoxMultipleModulations.setText(_translate("DialogOptions", "Enable modulation profiles"))
        self.groupBoxModulationAccuracy.setTitle(_translate("DialogOptions", "Modulation Accuracy"))
        self.radioButtonLowModulationAccuracy.setText(_translate("DialogOptions", "Low (2x8 bit) - Recommended for HackRF and RTL-SDR"))
        self.radioButtonMediumModulationAccuracy.setText(_translate("DialogOptions", "Medium (2x16 bit) - Recommended for BladeRF, PlutoSDR and SDRPlay"))
        self.radioButtonHighModulationAccuracy.setText(_translate("DialogOptions", "High (2x32 bit) - Recommended if you are not sure what to choose"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tabGeneration), _translate("DialogOptions", "Generation"))
        self.label_7.setText(_translate("DialogOptions", "Default View:"))
        self.comboBoxDefaultView.setItemText(0, _translate("DialogOptions", "Bit"))
        self.comboBoxDefaultView.setItemText(1, _translate("DialogOptions", "Hex"))
        self.comboBoxDefaultView.setItemText(2, _translate("DialogOptions", "ASCII"))
        self.checkBoxShowConfirmCloseDialog.setText(_translate("DialogOptions", "Show \"confirm close\" dialog"))
        self.checkBoxHoldShiftToDrag.setToolTip(_translate("DialogOptions", "<html><head/><body><p>If checked, you need to <span style=\" font-weight:600;\">hold the Shift key to drag</span> with the mouse inside graphic views like the drawn signal in Interpretation tab, while making a selection with the mouse does not require holding any buttons.</p><p>If unchecked, this is inverted: Hold shift to make a selection, and drag by default.</p></body></html>"))
        self.checkBoxHoldShiftToDrag.setText(_translate("DialogOptions", "Hold shift to drag"))
        self.checkBoxPauseTime.setText(_translate("DialogOptions", "Show pauses as time"))
        self.checkBoxAlignLabels.setText(_translate("DialogOptions", "Align on labels"))
        self.labelFontSize.setText(_translate("DialogOptions", "<html><head/><body><p>Application font size (<span style=\" font-weight:600;\">restart</span> for full effect):</p></body></html>"))
        self.doubleSpinBoxFontSize.setSuffix(_translate("DialogOptions", "pt"))
        self.label_9.setText(_translate("DialogOptions", "Choose application theme (requires restart):"))
        self.comboBoxTheme.setItemText(0, _translate("DialogOptions", "native look (default)"))
        self.comboBoxTheme.setItemText(1, _translate("DialogOptions", "fallback theme"))
        self.comboBoxTheme.setItemText(2, _translate("DialogOptions", "fallback theme (dark)"))
        self.labelIconTheme.setText(_translate("DialogOptions", "Choose icon theme (requires restart):"))
        self.comboBoxIconTheme.setItemText(0, _translate("DialogOptions", "bundled icons (default)"))
        self.comboBoxIconTheme.setItemText(1, _translate("DialogOptions", "native icon theme"))
        self.groupBoxSpectrogramColormap.setTitle(_translate("DialogOptions", "Spectrogram Colormap"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tabView), _translate("DialogOptions", "View"))
        self.btnAddLabelType.setText(_translate("DialogOptions", "..."))
        self.btnRemoveLabeltype.setText(_translate("DialogOptions", "..."))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tabFieldtypes), _translate("DialogOptions", "Fieldtypes"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab_plugins), _translate("DialogOptions", "Plugins"))
        self.labelInfoDeviceTable.setText(_translate("DialogOptions", "<html><head/><body><p>Use the checkboxes in the table below to choose device backends and enable or disable devices. Disabled devices will not show up in device related dialogs such as send or receive.</p></body></html>"))
        self.labelDeviceMissingInfo.setText(_translate("DialogOptions", "<html><head/><body><p>Missing a native backend? Perform a <a href=\"health_check\"><span style=\" text-decoration: underline; color:#0000ff;\">health check</span></a>! If GNU Radio backend is not available double check the GNU Radio settings below.</p></body></html>"))
        self.groupBoxNativeOptions.setTitle(_translate("DialogOptions", "Native options"))
        self.labelLibDirs.setText(_translate("DialogOptions", "Library directories:"))
        self.lineEditIncludeDirs.setPlaceholderText(_translate("DialogOptions", "Comma separated list of additional include directories"))
        self.labelNativeRebuildInfo.setText(_translate("DialogOptions", "You can rebuild the native device extensions here. This is useful, when you installed a device driver afterwards or your drivers are stored in an unusual location."))
        self.lineEditLibDirs.setPlaceholderText(_translate("DialogOptions", "Comma separated list of additional library directories"))
        self.labelIncludeDirs.setText(_translate("DialogOptions", "Include directories:"))
        self.btnViewBuildLog.setText(_translate("DialogOptions", "View log"))
        self.btnRebuildNative.setToolTip(_translate("DialogOptions", "<html><head/><body><p>Rebuild the native device extensions. You need to restart URH after this, to use new extensions.</p></body></html>"))
        self.btnRebuildNative.setText(_translate("DialogOptions", "Rebuild"))
        self.labelRebuildNativeStatus.setText(_translate("DialogOptions", "Rebuild <x> new device extensions. Please restart URH to use them."))
        self.groupBox_3.setTitle(_translate("DialogOptions", "GNU Radio options (optional)"))
        self.label.setToolTip(_translate("DialogOptions", "<html><head/><body><p>Choose a python interpreter which has access to your GNU Radio installation, that is, you can do <span style=\" font-weight:600;\">python -c &quot;import gnuradio&quot;</span> with it on the command line.</p></body></html>"))
        self.label.setText(_translate("DialogOptions", "Python Interpreter:"))
        self.lineEditGRPythonInterpreter.setToolTip(_translate("DialogOptions", "<html><head/><body><p>Choose a python interpreter which has access to your GNU Radio installation, that is, you can do <span style=\" font-weight:600;\">python -c &quot;import gnuradio&quot;</span> with it on the command line.</p></body></html>"))
        self.lineEditGRPythonInterpreter.setPlaceholderText(_translate("DialogOptions", "Enter python interpreter path e.g. /usr/bin/python"))
        self.btnChooseGRPythonInterpreter.setText(_translate("DialogOptions", "..."))
        self.label_8.setText(_translate("DialogOptions", "Default sending repititions:"))
        self.spinBoxNumSendingRepeats.setSpecialValueText(_translate("DialogOptions", "Infinite"))
        self.label_5.setText(_translate("DialogOptions", "Use this percentage of available RAM for buffer allocation:"))
        self.doubleSpinBoxRAMThreshold.setSuffix(_translate("DialogOptions", "%"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tabDevices), _translate("DialogOptions", "Device"))
from urh.ui.KillerDoubleSpinBox import KillerDoubleSpinBox



import numpy as np
from urh.dev.native.lib.climesdr cimport *
from libc.stdlib cimport malloc, free
# noinspection PyUnresolvedReferences
from cython.view cimport array as cvarray  # needed for converting of malloc array to python array

from urh.util.Logger import logger

cdef lms_device_t *_c_device
cdef lms_stream_t stream

cdef size_t CHANNEL = 0
cdef bool IS_TX = False

cpdef set_tx(bool is_tx):
    global IS_TX
    IS_TX = <bool>is_tx

cpdef bool get_tx():
    return IS_TX

cpdef set_channel(size_t channel):
    global CHANNEL
    CHANNEL = <size_t>channel
    return 0

cpdef size_t get_channel():
    return CHANNEL


cpdef list get_device_list():
    """
    Obtain a list of LMS devices attached to the system
    """
    cdef lms_info_str_t *dev_list = <lms_info_str_t *> malloc(256 * sizeof(lms_info_str_t))
    result = LMS_GetDeviceList(dev_list)
    if result > 0:
        return [dev_list[i].decode('UTF-8') for i in range(0, result)]
    else:
        return []

cpdef int open(str info=None):
    """
    Opens device specified by the provided ::lms_dev_info string
    This function should be used to open a device based upon the results of LMS_GetDeviceList()
    device should be initialized to NULL 
    :return 0 on success, (-1) on failure
    """
    cdef char*c_info
    if info is None:
        c_info = NULL
    else:
        info_byte_string = info.encode('UTF-8')
        c_info = <char *> info_byte_string

    return LMS_Open(&_c_device, c_info, NULL)

cpdef int close():
    """
    Close device
    :return:  0 on success, (-1) on failure
    """
    return LMS_Close(_c_device)

cpdef int init():
    """
    Configure LMS chip with settings that make it ready for operation.
 
    This configuration differs from default LMS chip configuration which is
    described in chip datasheet. In order to load default chip configuration use LMS_Reset().
    :return: 0 on success, (-1) on failure
    """
    return LMS_Init(_c_device)


cpdef int reset():
    return LMS_Reset(_c_device)

cpdef int synchronize():
    return LMS_Synchronize(_c_device, True)

cpdef int get_num_channels():
    """
    Obtain number of RX or TX channels. Use this to determine the maximum
    channel index (specifying channel index is required by most API functions).
    The maximum channel index is N-1, where N is number returned by this function
    :return:  Number of channels on success, (-1) on failure
    """
    return LMS_GetNumChannels(_c_device, IS_TX)

cpdef int enable_channel(bool enabled, bool is_tx, size_t channel):
    """
    Enable or disable specified RX channel.
    
    :param enabled: true(1) to enable, false(0) to disable.
    :return:  0 on success, (-1) on failure
    """
    return LMS_EnableChannel(_c_device, is_tx, channel, enabled)

cpdef int disable_current_channel():
    return enable_channel(False, IS_TX, CHANNEL)

cpdef enable_all_channels():
    enable_channel(True, False, 0)
    enable_channel(True, False, 1)
    enable_channel(True, True, 0)
    enable_channel(True, True, 1)

cpdef disable_all_channels():
    enable_channel(False, False, 0)
    enable_channel(False, False, 1)
    enable_channel(False, True, 0)
    enable_channel(False, True, 1)

cpdef int set_sample_rate(float_type rate, size_t oversample=0):
    """
    Set sampling rate for all RX/TX channels. Sample rate is in complex samples
    (1 sample = I + Q). The function sets sampling rate that is used for data
    exchange with the host. It also allows to specify higher sampling rate to be
    used in RF by setting oversampling ratio. Valid oversampling values are 1, 2,
    4, 8, 16, 32 or 0 (use device default oversampling value).
    :param rate: sampling rate in Hz to set
    :param oversample: RF oversampling ratio
    :return:  0 on success, (-1) on failure
    """
    LMS_SetSampleRate(_c_device, rate, oversample)

cpdef tuple get_sample_rate():
    """
    Get the sampling rate of the specified LMS device RX or TX channel.
    The function obtains the sample rate used in data interface with the host and
    the RF sample rate used by DAC/ADC.
    :return: tuple of host_Hz, rf_Hz or tuple -1,-1 on Error
    """
    cdef float_type host_hz = 0.0  # sampling rate used for data exchange with the host
    cdef float_type rf_hz = 0.0  # RF sampling rate in Hz

    result = LMS_GetSampleRate(_c_device, IS_TX, CHANNEL, &host_hz, &rf_hz)
    if result == 0:
        return host_hz, rf_hz
    else:
        return -1, -1

cpdef tuple get_sample_rate_range():
    """
    Get the range of supported sampling rates.
    :return: Tuple (start, end, step) of Allowed sample rate range in Hz, (-1, -1, -1) on Error
    """
    cdef lms_range_t sample_rate_range
    result = LMS_GetSampleRateRange(_c_device, IS_TX, &sample_rate_range)
    if result == 0:
        return sample_rate_range.min, sample_rate_range.max, sample_rate_range.step
    else:
        return -1, -1, -1

cpdef int set_center_frequency(float_type frequency):
    """
    Set RF center frequency in Hz. This automatically selects the appropriate
    antenna (band path) for the desired frequency. In order to override antenna selection use LMS_SetAntenna().
    :param frequency: Desired RF center frequency in Hz
    :return: 0 on success, (-1) on failure
    """
    return LMS_SetLOFrequency(_c_device, IS_TX, CHANNEL, frequency)

cpdef float_type get_center_frequency():
    """
    Obtain the current RF center frequency in Hz.
    
    :return: Current RF center frequency in Hz on success, (-1) on failure
    """
    cdef float_type frequency = 0.0
    result = LMS_GetLOFrequency(_c_device, IS_TX, CHANNEL, &frequency)
    if result == 0:
        return frequency
    else:
        return -1

cpdef tuple get_center_frequency_range():
    """
    Obtain the supported RF center frequency range in Hz.
    
    :return: Tuple (start, end, step) of allowed center freq range in Hz, (-1, -1, -1) on Error
    """
    cdef lms_range_t center_freq_range
    result = LMS_GetLOFrequencyRange(_c_device, IS_TX, &center_freq_range)
    if result == 0:
        return center_freq_range.min, center_freq_range.max, center_freq_range.step
    else:
        return -1, -1, -1

cpdef int set_normalized_gain(float_type gain):
    """
    Set the combined gain value
    
    This function computes and sets the optimal gain values of various amplifiers
    that are present in the device based on desired normalized gain value.
    
    :param gain: Desired gain, range [0, 1.0], where 1.0 represents the maximum gain
    :return:  0 on success, (-1) on failure
    """
    return LMS_SetNormalizedGain(_c_device, IS_TX, CHANNEL, gain)

cpdef float_type get_normalized_gain():
    """
    Obtain the current combined gain value
    
    :return: Current gain, range [0, 1.0], where 1.0 represents the maximum gain, or -1 on error
    """
    cdef float_type gain = 0.0
    result = LMS_GetNormalizedGain(_c_device, IS_TX, CHANNEL, &gain)
    if result == 0:
        return gain
    else:
        return -1

cpdef int set_lpf_bandwidth(float_type bandwidth):
    """
    Configure analog LPF of the LMS chip for the desired RF bandwidth.
    This function automatically enables LPF.
    
    :param bandwidth: LPF bandwidth in Hz
    :return: 0 on success, (-1) on failure
    """
    return LMS_SetLPFBW(_c_device, IS_TX, CHANNEL, bandwidth)

cpdef float_type get_lpf_bandwidth():
    """
    Get the currently configured analog LPF RF bandwidth.
    
    :return: Current LPF bandwidth in Hz on success, (-1) on failure
    """
    cdef float_type bandwidth = 0.0
    result = LMS_GetLPFBW(_c_device, IS_TX, CHANNEL, &bandwidth)
    if result == 0:
        return bandwidth
    else:
        return -1

cpdef get_lpf_bandwidth_range():
    """
    Get the RF bandwidth setting range supported by the analog LPF of LMS chip
    
    :return: Tuple (start, end, step) of allowed bandwidth values in Hz, (-1, -1, -1) on Error
    """
    cdef lms_range_t bandwidth_range
    result = LMS_GetLPFBWRange(_c_device, IS_TX, &bandwidth_range)
    if result == 0:
        return bandwidth_range.min, bandwidth_range.max, bandwidth_range.step
    else:
        return -1, -1, -1

cpdef calibrate(double bw):
    """
    Perform the automatic calibration of specified RX/TX channel. The automatic
    calibration must be run after device configuration is finished because
    calibration values are dependent on various configuration settings.

    automatic RX calibration is not available when RX_LNA_H path is
    selected

    Device should be configured
    
    :param bw: bandwidth
    :return: 0 on success, (-1) on failure
    """
    return LMS_Calibrate(_c_device, IS_TX, CHANNEL, bw, 0)

cpdef list get_antenna_list():
    """
    Obtain antenna list with names. First item in the list is the name of antenna index 0.
    :return: 
    """
    cdef lms_name_t *ant_list = <lms_name_t *> malloc(256 * sizeof(lms_name_t))
    result = LMS_GetAntennaList(_c_device, IS_TX, CHANNEL, ant_list)
    if result > 0:
        return [ant_list[i].decode('UTF-8') for i in range(0, result)]
    else:
        return []

cpdef int set_antenna(size_t index):
    """
    Select the antenna for the specified RX or TX channel.
    
    LMS_SetFrequency() automatically selects antenna based on frequency. 
    This function is meant to override path selected by LMS_SetFrequency() and should be called after LMS_SetFrequency().
    :param index: Index of antenna to select
    :return: 0 on success, (-1) on failure
    """
    return LMS_SetAntenna(_c_device, IS_TX, CHANNEL, index)

cpdef int get_antenna():
    """
    Obtain currently selected antenna of the the specified RX or TX channel.
     
    :return: Index of selected antenna on success, (-1) on failure
    """
    return LMS_GetAntenna(_c_device, IS_TX, CHANNEL)

cpdef tuple get_antenna_bw(size_t index):
    """
    Obtains bandwidth (lower and upper frequency) of the specified antenna

    :param index: Antenna index
    :return: Tuple (start, end, step) of allowed bandwidth values in Hz, (-1, -1, -1) on Error
    """
    cdef lms_range_t bandwidth_range
    result = LMS_GetAntennaBW(_c_device, IS_TX, CHANNEL, index, &bandwidth_range)
    if result == 0:
        return bandwidth_range.min, bandwidth_range.max, bandwidth_range.step
    else:
        return -1, -1, -1

cpdef tuple get_nco_frequency():
    cdef float_type freq = 0.0
    cdef float_type pho = 0.0
    result = LMS_GetNCOFrequency(_c_device, IS_TX, CHANNEL, &freq, &pho)
    if result == 0:
        return freq, pho
    else:
        return -1, 1

cpdef float_type get_clock_freq(size_t clk_id):
    cdef float_type clock_hz = 0.0
    result = LMS_GetClockFreq(_c_device, clk_id, &clock_hz)
    if result == 0:
        return clock_hz
    else:
        return -1

cpdef int set_clock_freq(size_t clk_id, float_type frequency_hz):
    return LMS_SetClockFreq(_c_device, clk_id, frequency_hz)

cpdef float_type get_chip_temperature():
    """
    Read LMS7 chip internal temperature sensor
    
    :return: Temperature on success, (-1) on failure
    """
    cdef float_type chip_temp = 0.0
    result = LMS_GetChipTemperature(_c_device, 0, &chip_temp)
    if result == 0:
        return chip_temp
    else:
        return -1

cpdef int setup_stream(uint32_t fifo_size):
    """
    Create new stream based on parameters passed in configuration structure.
    The structure is initialized with stream handle.
    :param fifo_size: FIFO size (in samples) used by stream.
    :return: 0 on success, (-1) on failure
    """
    stream.isTx = IS_TX
    stream.channel = <uint32_t> CHANNEL
    stream.fifoSize = fifo_size
    stream.dataFmt = dataFmt_t.LMS_FMT_F32
    stream.throughputVsLatency = 0.0  # optimize for minimum latency

    return LMS_SetupStream(_c_device, &stream)

cpdef int destroy_stream():
    """
    Deallocate memory used for stream.
    :return: 0 on success, (-1) on failure
    """
    LMS_DestroyStream(_c_device, &stream)

cpdef int start_stream():
    """
    Start stream 
    :return: 0 on success, (-1) on failure
    """
    return LMS_StartStream(&stream)

cpdef int stop_stream():
    """
    Stop stream
    :return: 0 on success, (-1) on failure
    """
    return LMS_StopStream(&stream)

cpdef int recv_stream(connection, unsigned num_samples, unsigned timeout_ms):
    """
    Read samples from the FIFO of the specified stream.
    Sample buffer must be big enough to hold requested number of samples.
    
    :param num_samples: how many samples shall be read from streams FIFO
    :param connection: multiprocessing connection to send the received samples to
    :param timeout_ms: how long to wait for data before timing out.
    :return: 
    """
    cdef lms_stream_meta_t meta = lms_stream_meta_t(0, False, False)
    cdef float*buff = <float *> malloc(num_samples * 2 * sizeof(float))

    if not buff:
        raise MemoryError()

    cdef int received_samples = LMS_RecvStream(&stream, buff, num_samples, &meta, timeout_ms)

    if received_samples > 0:
        connection.send_bytes(<float[:2*received_samples]>buff)
    else:
        logger.warning("LimeSDR: Failed to receive stream")

    free(buff)

cpdef int send_stream(float[::1] samples, unsigned timeout_ms):
    """
    Write samples to the FIFO of the specified stream.
    
    :param samples: sample buffer
    :param timeout_ms: how long to wait for data before timing out
    :return: number of samples send on success, (-1) on failure
    """
    cdef lms_stream_meta_t meta = lms_stream_meta_t(0, False, False)
    if len(samples) == 1:
        samples = np.zeros(1020, dtype=np.float32)
    cdef size_t sample_count = len(samples) // 2

    if len(samples) > 0:
        return LMS_SendStream(&stream, &samples[0], sample_count, &meta, timeout_ms)
    else:
        return -1

cpdef load_config(filename):
    filename_byte_string = filename.encode('UTF-8')
    c_filename = <char *> filename_byte_string
    LMS_LoadConfig(_c_device, c_filename)

cpdef save_config(str filename):
    filename_byte_string = filename.encode('UTF-8')
    c_filename = <char *> filename_byte_string
    LMS_SaveConfig(_c_device, c_filename)

cpdef void print_last_error():
    cdef char * error_msg = <char *> malloc(2000 * sizeof(char))
    error_msg = <char *>LMS_GetLastErrorMessage()
    error_msg_py = error_msg.decode("UTF-8")
    print(error_msg_py)



import array
import datetime
import re
import threading
import time

import numpy
from PyQt5.QtCore import pyqtSignal, QObject, pyqtSlot
from PyQt5.QtTest import QSignalSpy

from urh.dev.BackendHandler import BackendHandler, Backends
from urh.dev.EndlessSender import EndlessSender
from urh.signalprocessing.ChecksumLabel import ChecksumLabel
from urh.signalprocessing.Message import Message
from urh.signalprocessing.Modulator import Modulator
from urh.signalprocessing.Participant import Participant
from urh.signalprocessing.ProtocolSniffer import ProtocolSniffer
from urh.simulator.SimulatorConfiguration import SimulatorConfiguration
from urh.simulator.SimulatorCounterAction import SimulatorCounterAction
from urh.simulator.SimulatorExpressionParser import SimulatorExpressionParser
from urh.simulator.SimulatorGotoAction import SimulatorGotoAction
from urh.simulator.SimulatorMessage import SimulatorMessage
from urh.simulator.SimulatorProtocolLabel import SimulatorProtocolLabel
from urh.simulator.SimulatorRule import SimulatorRule, SimulatorRuleCondition, ConditionType
from urh.simulator.SimulatorSleepAction import SimulatorSleepAction
from urh.simulator.SimulatorTriggerCommandAction import SimulatorTriggerCommandAction
from urh.simulator.Transcript import Transcript
from urh.util import util, HTMLFormatter
from urh.util.Logger import logger
from urh.util.ProjectManager import ProjectManager


class Simulator(QObject):
    simulation_started = pyqtSignal()
    simulation_stopped = pyqtSignal()

    def __init__(self, simulator_config: SimulatorConfiguration, modulators,
                 expression_parser: SimulatorExpressionParser, project_manager: ProjectManager,
                 sniffer: ProtocolSniffer, sender: EndlessSender):
        super().__init__()
        self.simulator_config = simulator_config
        self.project_manager = project_manager
        self.expression_parser = expression_parser
        self.modulators = modulators  # type: list[Modulator]
        self.backend_handler = BackendHandler()

        self.transcript = Transcript()

        self.current_item = None
        self.last_sent_message = None
        self.is_simulating = False
        self.do_restart = False
        self.current_repeat = 0
        self.log_messages = []

        self.sniffer_ready = False
        self.sender_ready = False
        self.fatal_device_error_occurred = False

        self.verbose = True

        self.sniffer = sniffer
        self.sender = sender

    def __initialize_counters(self):
        for item in self.simulator_config.get_all_items():
            if isinstance(item, SimulatorCounterAction):
                item.reset_value()

    def start(self):
        self.reset()

        self.transcript.clear()
        self.__initialize_counters()

        # start devices
        if self.sniffer:
            self.sniffer.rcv_device.fatal_error_occurred.connect(self.stop_on_error)
            self.sniffer.rcv_device.ready_for_action.connect(self.on_sniffer_ready)

        if self.sender:
            self.sender.device.fatal_error_occurred.connect(self.stop_on_error)
            self.sender.device.ready_for_action.connect(self.on_sender_ready)

        if self.sniffer:
            self.sniffer.sniff()

        if self.sender:
            self.sender.start()

        self._start_simulation_thread()

        # Ensure all ongoing qt signals can be processed
        time.sleep(0.1)

    @pyqtSlot(str)
    def stop_on_error(self, msg: str):
        self.fatal_device_error_occurred = True
        if self.is_simulating:
            self.stop(msg=msg)

    @pyqtSlot()
    def on_sniffer_ready(self):
        if not self.sniffer_ready:
            self.log_message("RX is ready to operate")
            self.sniffer_ready = True

    @pyqtSlot()
    def on_sender_ready(self):
        if not self.sender_ready:
            self.log_message("TX is ready to operate")
            self.sender_ready = True

    def stop(self, msg=""):
        self.simulation_stopped.emit()

        if self.is_simulating:
            self.log_message("Stop simulation" + (" ({})".format(msg.strip()) if msg else ""))
            self.is_simulating = False
            self.do_restart = False
            self.simulation_thread.join(2.5)

        # stop devices
        if self.sniffer:
            self.sniffer.stop()

        if self.sender:
            self.sender.stop()

    def restart(self):
        self.transcript.start_new_round()
        self.reset()
        self.log_message("<b>Restarting simulation</b>")

    def reset(self):
        self.sniffer_ready = False
        self.sender_ready = False
        self.fatal_device_error_occurred = False

        if self.sniffer:
            self.sniffer.clear()

        self.current_item = self.simulator_config.rootItem

        for msg in self.simulator_config.get_all_messages():
            msg.send_recv_messages[:] = []

        self.last_sent_message = None
        self.is_simulating = True
        self.do_restart = False
        self.current_repeat = 0
        self.log_messages[:] = []

    @property
    def devices(self):
        result = []
        if self.sniffer is not None:
            result.append(self.sniffer.rcv_device)
        if self.sender is not None:
            result.append(self.sender.device)
        return result

    def device_messages(self) -> list:
        return [device.read_messages() for device in self.devices]

    def read_log_messages(self):
        result = self.log_messages[:]
        self.log_messages.clear()
        return result

    def cleanup(self):
        for device in self.devices:
            if device.backend not in (Backends.none, Backends.network):
                device.cleanup()

            if device is not None:
                device.free_data()

    def _start_simulation_thread(self):
        self.simulation_thread = threading.Thread(target=self.simulate)
        self.simulation_thread.daemon = True
        self.simulation_thread.start()

    def simulation_is_finished(self):
        if self.project_manager.simulator_num_repeat == 0:
            return False

        return self.current_repeat >= self.project_manager.simulator_num_repeat

    def __wait_for_devices(self):
        for i in range(10):
            if (self.sniffer is None or self.sniffer_ready) and (self.sender is None or self.sender_ready):
                return True
            if self.fatal_device_error_occurred:
                return False
            self.log_message("<i>Waiting for devices</i>")
            time.sleep(1)

        return True

    def __fill_counter_values(self, command: str):
        result = []
        regex = r"(item[0-9]+\.counter_value)"
        for token in re.split(regex, command):
            if re.match(regex, token) is not None:
                try:
                    result.append(str(self.simulator_config.item_dict[token].value))
                except (KeyError, ValueError, AttributeError):
                    logger.error("Could not get counter value for " + token)
            else:
                result.append(token)

        return "".join(result)

    def simulate(self):
        self.simulation_started.emit()
        self.is_simulating = self.__wait_for_devices()

        if not self.is_simulating:
            # Simulation may have ended due to device errors
            self.stop("Devices not ready")
            return

        self.log_message("<b>Simulation is running</b>")

        while self.is_simulating and not self.simulation_is_finished():
            if self.current_item is self.simulator_config.rootItem:
                next_item = self.current_item.next()

            elif isinstance(self.current_item, SimulatorProtocolLabel):
                next_item = self.current_item.next()

            elif isinstance(self.current_item, SimulatorMessage):
                self.process_message()
                next_item = self.current_item.next()

            elif isinstance(self.current_item, SimulatorGotoAction):
                next_item = self.current_item.target
                self.log_message("GOTO item " + next_item.index())

            elif isinstance(self.current_item, SimulatorTriggerCommandAction):
                next_item = self.current_item.next()
                command = self.__fill_counter_values(self.current_item.command)
                self.log_message("Calling {}".format(command))
                if self.current_item.pass_transcript:
                    transcript = "\n".join(self.transcript.get_for_all_participants(all_rounds=False))
                    result, rc = util.run_command(command, transcript, use_stdin=True, return_rc=True)
                else:
                    result, rc = util.run_command(command, param=None, detailed_output=True, return_rc=True)
                self.current_item.return_code = rc
                self.log_message(result)

            elif isinstance(self.current_item, SimulatorRule):
                condition = self.current_item.get_first_applying_condition()

                if condition is not None and condition.logging_active and condition.type != ConditionType.ELSE:
                    self.log_message("Rule condition " + condition.index() + " (" + condition.condition + ") applied")

                if condition is not None and condition.child_count() > 0:
                    next_item = condition.children[0]
                else:
                    next_item = self.current_item.next_sibling()

            elif isinstance(self.current_item, SimulatorRuleCondition):
                if self.current_item.type == ConditionType.IF:
                    next_item = self.current_item.parent()
                else:
                    next_item = self.current_item.parent().next_sibling()

            elif isinstance(self.current_item, SimulatorSleepAction):
                self.log_message(self.current_item.caption)
                time.sleep(self.current_item.sleep_time)
                next_item = self.current_item.next()

            elif isinstance(self.current_item, SimulatorCounterAction):
                self.current_item.progress_value()
                self.log_message("Increase counter by {} to {}".format(self.current_item.step, self.current_item.value))
                next_item = self.current_item.next()

            elif self.current_item is None:
                self.current_repeat += 1
                next_item = self.simulator_config.rootItem
                self.transcript.start_new_round()

            else:
                raise ValueError("Unknown action {}".format(type(self.current_item)))

            self.current_item = next_item

            if self.do_restart:
                self.restart()

        self.stop(msg="Finished")

    def process_message(self):
        assert isinstance(self.current_item, SimulatorMessage)
        msg = self.current_item

        if msg.source is None:
            return

        new_message = self.generate_message_from_template(msg)

        if msg.source.simulate:
            # we have to send a message
            sender = self.sender
            if sender is None:
                self.log_message("Fatal: No sender configured")
                return

            for lbl in new_message.message_type:
                if isinstance(lbl.label, ChecksumLabel):
                    checksum = lbl.label.calculate_checksum_for_message(new_message, use_decoded_bits=False)
                    label_range = new_message.get_label_range(lbl=lbl.label, view=0, decode=False)
                    start, end = label_range[0], label_range[1]
                    new_message.plain_bits[start:end] = checksum + array.array("B", [0] * (
                            (end - start) - len(checksum)))

            self.transcript.append(msg.source, msg.destination, new_message, msg.index())
            self.send_message(new_message, msg.repeat, sender, msg.modulator_index)
            self.log_message("Sending message " + msg.index())
            self.log_message_labels(new_message)
            msg.send_recv_messages.append(new_message)
            self.last_sent_message = msg
        else:
            # we have to receive a message
            self.log_message("Waiting for message {}...".format(msg.index()))
            sniffer = self.sniffer
            if sniffer is None:
                self.log_message("Fatal: No sniffer configured")
                return

            retry = 0

            max_retries = self.project_manager.simulator_retries
            while self.is_simulating and not self.simulation_is_finished() and retry < max_retries:
                received_msg = self.receive_message(sniffer)
                self.log_message("  Received {} data bits".format(len(received_msg)))

                if not self.is_simulating:
                    return

                if received_msg is None:
                    if self.project_manager.simulator_error_handling_index == 0:
                        self.resend_last_message()
                        retry += 1
                        continue
                    elif self.project_manager.simulator_error_handling_index == 1:
                        self.stop()
                        return
                    elif self.project_manager.simulator_error_handling_index == 2:
                        self.do_restart = True
                        return

                received_msg.decoder = new_message.decoder
                received_msg.message_type = new_message.message_type

                self.log_message("  Check whether received data matches")
                check_result, error_msg = self.check_message(received_msg, new_message, retry=retry,
                                                             msg_index=msg.index())

                if check_result:
                    decoded_msg = Message(received_msg.decoded_bits, 0,
                                          received_msg.message_type, decoder=received_msg.decoder)
                    msg.send_recv_messages.append(decoded_msg)
                    self.transcript.append(msg.source, msg.destination, decoded_msg, msg.index())
                    self.log_message("Received message " + msg.index() + ": ")
                    self.log_message_labels(decoded_msg)
                    return
                elif self.verbose:
                    self.log_message(error_msg)

                retry += 1

            if retry == self.project_manager.simulator_retries:
                self.log_message("Message " + msg.index() + " not received")
                self.stop()

    def log_message(self, message):
        timestamp = '{0:%b} {0.day} {0:%H}:{0:%M}:{0:%S}.{0:%f}'.format(datetime.datetime.now())

        if isinstance(message, list) and len(message) > 0:
            self.log_messages.append(timestamp + ": " + message[0])
            self.log_messages.extend(message[1:])
            logger.debug("\n".join(message))
        else:
            self.log_messages.append(timestamp + ": " + message)
            logger.debug(message)

    def check_message(self, received_msg, expected_msg, retry: int, msg_index: int) -> (bool, str):
        if len(received_msg.decoded_bits) == 0:
            return False, "Failed to decode message {}".format(msg_index)

        for lbl in received_msg.message_type:
            if lbl.value_type_index in (1, 4):
                # get live, random
                continue

            start_recv, end_recv = received_msg.get_label_range(lbl.label, 0, True)
            start_exp, end_exp = expected_msg.get_label_range(lbl.label, 0, False)

            if isinstance(lbl.label, ChecksumLabel):
                expected = lbl.label.calculate_checksum_for_message(received_msg, use_decoded_bits=True)
                start, end = received_msg.get_label_range(lbl.label, 0, True)
                actual = received_msg.decoded_bits[start:end]
            else:
                actual = received_msg.decoded_bits[start_recv:end_recv]
                expected = expected_msg[start_exp:end_exp]

            if actual != expected:
                log_msg = []
                log_msg.append("Attempt for message {} [{}/{}]".format(msg_index, retry + 1,
                                                                       self.project_manager.simulator_retries))
                log_msg.append(HTMLFormatter.indent_string("Mismatch for label: <b>{}</b>".format(lbl.name)))
                expected_str = util.convert_bits_to_string(expected, lbl.label.display_format_index)
                got_str = util.convert_bits_to_string(actual, lbl.label.display_format_index)
                log_msg.append(HTMLFormatter.align_expected_and_got_value(expected_str, got_str, align_depth=2))
                return False, log_msg

        return True, ""

    def log_message_labels(self, message: Message):
        message.split(decode=False)
        for lbl in message.message_type:
            if not lbl.logging_active:
                continue

            try:
                data = message.plain_bits[lbl.start:lbl.end]
            except IndexError:
                return None

            lsb = lbl.display_bit_order_index == 1
            lsd = lbl.display_bit_order_index == 2

            data = util.convert_bits_to_string(data, lbl.display_format_index, pad_zeros=True, lsb=lsb, lsd=lsd)
            if data is None:
                continue

            log_msg = lbl.name + ": " + HTMLFormatter.monospace(data)
            self.log_messages.append(HTMLFormatter.indent_string(log_msg))

    def resend_last_message(self):
        self.log_message("Resending last message")
        lsm = self.last_sent_message

        if lsm is None:
            return

        sender = self.sender
        self.send_message(lsm.send_recv_messages[-1], lsm.repeat, sender, lsm.modulator_index)

    def send_message(self, message, repeat, sender, modulator_index):
        modulator = self.modulators[modulator_index]
        modulated = modulator.modulate(message.encoded_bits, pause=message.pause, dtype=self.sender.device.data_type)

        curr_repeat = 0

        while curr_repeat < repeat:
            sender.push_data(modulated)
            curr_repeat += 1

    def receive_message(self, sniffer):
        if len(sniffer.messages) > 0:
            return sniffer.messages.pop(0)

        if QSignalSpy(sniffer.message_sniffed).wait(self.project_manager.simulator_timeout_ms):
            try:
                return sniffer.messages.pop(0)
            except IndexError:
                self.log_message("Could not receive message")
                return None
        else:
            self.log_message("Receive timeout")
            return None

    def get_full_transcript(self, start=0, use_bit=True):
        result = []
        for source, destination, msg, msg_index in self.transcript[start:]:
            try:
                data = msg.plain_bits_str if use_bit else msg.plain_hex_str
                result.append(self.TRANSCRIPT_FORMAT.format(msg_index, source.shortname, destination.shortname, data))
            except AttributeError:
                result.append("")
        return result

    def generate_message_from_template(self, template_msg: SimulatorMessage):
        new_message = Message(template_msg.plain_bits, pause=template_msg.pause, rssi=0,
                              message_type=template_msg.message_type, decoder=template_msg.decoder)

        for lbl in template_msg.children:  # type: SimulatorProtocolLabel
            if lbl.value_type_index == 2:
                # formula
                valid, _, node = self.expression_parser.validate_expression(lbl.formula)
                assert valid
                result = self.expression_parser.evaluate_node(node)
            elif lbl.value_type_index == 3:
                transcript = self.transcript.get_for_participant(template_msg.source
                                                                 if template_msg.source.simulate
                                                                 else template_msg.destination)

                if template_msg.destination.simulate:
                    direction = "->" if template_msg.source.simulate else "<-"
                    transcript += "\n" + direction + new_message.plain_bits_str + "\n"

                cmd = self.__fill_counter_values(lbl.external_program)
                result = util.run_command(cmd, transcript, use_stdin=True)
                if len(result) != lbl.end - lbl.start:
                    log_msg = "Result value of external program {}: {} ({}) does not match label length {}"
                    logger.error(log_msg.format(cmd, result, len(result), lbl.end - lbl.start))
                    continue

                try:
                    new_message[lbl.start:lbl.end] = array.array("B", (map(bool, map(int, result))))
                except Exception as e:
                    log_msg = "Could not assign {} to range because {}".format(result, e)
                    logger.error(log_msg)

                continue
            elif lbl.value_type_index == 4:
                # random value
                result = numpy.random.randint(lbl.random_min, lbl.random_max + 1)
            else:
                continue

            self.set_label_value(new_message, lbl, result)

        return new_message

    @staticmethod
    def set_label_value(message, label, decimal_value: int):
        lbl_len = label.end - label.start
        f_string = "{0:0" + str(lbl_len) + "b}"
        bits = f_string.format(decimal_value)

        if len(bits) > lbl_len:
            logger.warning("Value {0} too big for label {1}, bits truncated".format(decimal_value, label.name))

        for i in range(lbl_len):
            message[label.start + i] = bool(int(bits[i]))



from subprocess import call, Popen

MODULES = ["path_creator", "signal_functions", "util", "auto_interpretation"]
COMPILER_DIRECTIVES = {'language_level': 3,
                       'cdivision': True,
                       'wraparound': False,
                       'boundscheck': False,
                       'initializedcheck': False,
                       }

for module in MODULES:
    call(["cython", "-a", "-X", ",".join("{}={}".format(key, val) for key, val in COMPILER_DIRECTIVES.items()),
          "--cplus", "-3", module + ".pyx"])
    Popen(["firefox", module + ".html"])



from collections import defaultdict

from PyQt5.QtCore import pyqtSignal, QModelIndex, Qt

from urh import settings
from urh.models.TableModel import TableModel
from urh.signalprocessing.ProtocolAnalyzer import ProtocolAnalyzer
from urh.ui.actions.DeleteBitsAndPauses import DeleteBitsAndPauses


class ProtocolTableModel(TableModel):
    ref_index_changed = pyqtSignal(int)

    def __init__(self, proto_analyzer: ProtocolAnalyzer, participants, controller, parent=None):
        super().__init__(participants=participants, parent=parent)

        self.controller = controller # type: urh.controller.CompareFrameController.CompareFrameController

        self.protocol = proto_analyzer
        self.active_group_ids = [0]

    @property
    def diff_columns(self) -> defaultdict(set):
        return self._diffs

    @property
    def refindex(self):
        return self._refindex

    @refindex.setter
    def refindex(self, refindex):
        if refindex != self._refindex:
            self._refindex = refindex
            self.update()
            self.ref_index_changed.emit(self._refindex)

    def refresh_fonts(self):
        self.bold_fonts.clear()
        self.text_colors.clear()
        for i in self._diffs.keys():
            for j in self._diffs[i]:
                self.bold_fonts[i, j] = True
                self.text_colors[i, j] = settings.DIFFERENCE_CELL_COLOR

        if self._refindex >= 0:
            for j in range(self.col_count):
                self.text_colors[self._refindex, j] = settings.SELECTED_ROW_COLOR

    def delete_range(self, min_row: int, max_row: int, start: int, end: int):
        if not self.is_writeable:
            return

        del_action = DeleteBitsAndPauses(proto_analyzer=self.protocol, start_message=min_row, end_message=max_row,
                                         start=start, end=end, view=self.proto_view, decoded=True,
                                         subprotos=self.controller.protocol_list, update_label_ranges=False)
        self.undo_stack.push(del_action)

    def flags(self, index: QModelIndex):
        if index.isValid():
            alignment_offset = self.get_alignment_offset_at(index.row())
            if index.column() < alignment_offset:
                return Qt.ItemIsSelectable | Qt.ItemIsEnabled

            if self.is_writeable:
                return Qt.ItemIsEnabled | Qt.ItemIsEditable | Qt.ItemIsSelectable
            else:
                return Qt.ItemIsEnabled | Qt.ItemIsSelectable
        else:
            return Qt.NoItemFlags
